
// cplex.cpp - produce CPLEX integer linear program
// (c) 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include <fstream>
#include "graph.hpp"

#define TRIANGLE  // add triangle inequalities to specification

int main(int argc, char** argv)
{
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> "
            "<edgefeatures> < <params> > <cplexfile>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];
    
    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read graph from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    Isinf::Graph g(arg1);
    assert(g.is_connected());

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(3);
    }

    std::vector<double> f;
    std::vector<double> e;

    while (arg2.good())
    {
        while (e.size() < g.nedges)
        {
            arg2 >> f;
            if (e.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << e.size() << std::endl;
                exit(4);
            }
            e.push_back(f*theta);
            f.clear();
        }

        // output CPLEX file
        std::cout << "\nminimize E :";
        std::cout << std::showpos;
        for (size_t i = 0; i < e.size(); ++i)
        {
            const size_t a = std::min(g.edges[i]->node,
                                      g.edges[i]->other->node);
            const size_t b = std::max(g.edges[i]->node,
                                      g.edges[i]->other->node);
            std::cout << "\n    " << e[i] << " mu01(" << a << ',' << b
                << ") " << e[i] << " mu10(" << a << ',' << b << ')';
        }

        std::cout << "\n\nsubject to";
        for (size_t i = 0; i < e.size(); ++i)
        {
            const size_t a = std::min(g.edges[i]->node,
                                      g.edges[i]->other->node);
            const size_t b = std::max(g.edges[i]->node,
                                      g.edges[i]->other->node);

            // marginalisation constraints
            std::cout << "\n    mu00(" << a << ',' << b << ") + mu01("
                << a << ',' << b << ") - mu0(" << a << ") = 0";
            std::cout << "\n    mu10(" << a << ',' << b << ") + mu11("
                << a << ',' << b << ") + mu0(" << a << ") = 1";
            std::cout << "\n    mu00(" << a << ',' << b << ") + mu10("
                << a << ',' << b << ") - mu0(" << b << ") = 0";
            std::cout << "\n    mu01(" << a << ',' << b << ") + mu11("
                << a << ',' << b << ") + mu0(" << b << ") = 1";

#ifdef TRIANGLE
            // relate marginals to triangle
            std::cout << "\n    mu00(" << a << ',' << b << ") + mu11("
                << a << ',' << b << ") - t(" << a << ',' << b << ") = 0";
        }

        // triangle inequalities
        for (size_t i = 0; i < g.nodes.size(); ++i)
        {
            for (size_t j = i + 1; j < g.nodes.size(); ++j)
            {
                for (size_t k = j + 1; k < g.nodes.size(); ++k)
                {
                    std::cout << "\n    t(" << i << ',' << j << ") + t(" << j
                        << ',' << k << ") - t(" << i << ',' << k << ") <= 1";
                    std::cout << "\n    t(" << i << ',' << j << ") + t(" << j
                        << ',' << k << ") + t(" << i << ',' << k << ") >= 1";
                }
            }
#endif
        }

        std::cout << "\n\nbounds\n    mu0(" << g.nodes.size() - 1 << ") = 1";

        std::cout << "\n\nbinary";
        for (size_t i = 0; i < g.nodes.size(); ++i)
            std::cout << "\n    mu0(" << i << ')';

        for (size_t i = 0; i < e.size(); ++i)
        {
            const size_t a = std::min(g.edges[i]->node,
                                      g.edges[i]->other->node);
            const size_t b = std::max(g.edges[i]->node,
                                      g.edges[i]->other->node);
            std::cout << "\n    mu00(" << a << ',' << b << ')'
                << "\n    mu01(" << a << ',' << b << ')'
                << "\n    mu10(" << a << ',' << b << ')'
                << "\n    mu11(" << a << ',' << b << ')';
        }

        std::cout << "\n\nend\n\n";

        e.clear();
    }
}

