
// embed.cpp - planar embedder, uses John Boyer's planarity code
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include "utils.hpp"
#include "../planarity/graph.h"

#include <fstream>
#include <string>
#include <map>

typedef std::pair<size_t,size_t> Edge;

// set up map from node pair to edge index
typedef std::pair<Edge, size_t> EdgeMapEntry;
typedef std::map<Edge, size_t> EdgeMap;
EdgeMap edgemap;

// write out adjacency list of edge indices
void writeAdjList(graphP g)
{
     int i, j;

     for (i = 0; i < g->N; ++i)
     {
          j = g->G[i].link[1];
          while (j >= g->N)
          {
              Edge e(std::min(i, g->G[j].v), std::max(i, g->G[j].v));

              EdgeMap::iterator it(edgemap.find(e));
              if (it == edgemap.end())
              {
                  std::cerr << Isinf::argv0 << "can't find edge ("
                      << i << ',' << g->G[j].v << ')' << std::endl;
                  exit(5);
              }
              std::cout << it->second << ' ';

              j = g->G[j].link[1];
          }
          std::cout << std::endl;
     }
}

// handle planarity return codes
inline void errCheck(int k)
{
    if (k == OK) return;
    std::cerr << Isinf::argv0;
    switch (k)
    {
        case NOTOK:
            std::cerr << "something is wrong";
            break;
        case NONPLANAR:
            std::cerr << "graph is not planar";
            break;
        case NONOUTERPLANAR:
            std::cerr << "graph is not outerplanar";
            break;
    }
    std::cerr << std::endl;
    exit(k);
}

int main(int argc, char** argv)
{
    std::vector<Edge> edges;
    std::string s;
    size_t max = 0;
    std::filebuf* fb;

    Isinf::argv0 = argv[0];
    if (argc > 2)
    {
        std::cerr << "usage: " << argv[0]
            << " [<] <edgelist> > <embedding>" << std::endl;
        exit(1);
    }
    else if (argc == 2)
    {
        fb = new std::filebuf;
        if (fb->open(argv[1], std::ios::in) == NULL)
        {
            std::cerr << Isinf::argv0 << "can't open edge list '"
                << argv[1] << "' for input" << std::endl;
            exit(2);
        }
    }
    else fb = (std::filebuf*) std::cin.rdbuf();
    std::istream input(fb);

    // read edge list from input

    std::vector<size_t> x;
    int n = 0;
    input >> x;
    while (x.size() >= 2)
    {
        if (x[0] == x[1])  // self-loop!
        {
            std::cerr << Isinf::argv0 << "error at node " << x[0]
                << ": self-loops not allowed" << std::endl;
            exit(3);
        }

        if (x[0] > x[1]) std::swap(x[0], x[1]);
        if (x[1] > max) max = x[1];

        Edge e(x[0], x[1]);
        edges.push_back(e);

        if (!edgemap.insert(EdgeMapEntry(e, n++)).second)
        {
            std::cerr << Isinf::argv0 << "duplicate edge between nodes "
                << x[0] << " and " << x[1] << std::endl;
            exit(4);
        }

        x.clear();
        input >> x;
    }

    if (edges.size() < 2)
    {
        std::cerr << Isinf::argv0 << "less than 2 valid edges found" << std::endl;
        exit(2);
    }
    else if (edges.size() > 3*(++max) - 6)
        // edge limit for planar graphs
        errCheck(NONPLANAR);

    // create graph in planarity code
    graphP g = gp_New();
    gp_InitGraph(g, max);
    for (std::vector<Edge>::iterator i = edges.begin(); i < edges.end(); ++i)
        errCheck(gp_AddEdge(g, i->first, 0, i->second, 0));

    // timed call to planarity code
    Isinf::CPUtimer *t = new Isinf::CPUtimer;
    n = gp_Embed(g, EMBEDFLAGS_PLANAR);
    delete t;

    // report result
    if (n == OK)
    {
        gp_SortVertices(g);
        writeAdjList(g);  // output planar embedding
        gp_Free(&g);
        exit(n);
    }
    else
    {
        gp_Free(&g);
        errCheck(n);
    }
}

