
// evaluate.cpp - compare MAP or M3P states to target labels
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include "utils.hpp"

#include <iomanip>
#include <fstream>
#include <string>
#include <map>

typedef std::pair<size_t,size_t> Edge;

int main(int argc, char** argv)
{
    std::vector<Edge> edges;
    std::vector<size_t> x;
    size_t max = 0;

    Isinf::argv0 = argv[0];
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0]
            << " <edgelist> <nodestates> < <nodestates>" << std::endl;
        exit(1);
    }

    // read edge list from argv[1]
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    arg1 >> x;
    while (x.size() >= 2)
    {
        if (x[0] == x[1])  // self-loop!
        {
            std::cerr << Isinf::argv0 << "error at node " << x[0]
                << ": self-loops not allowed" << std::endl;
            exit(3);
        }

        if (x[0] > x[1]) std::swap(x[0], x[1]);
        if (x[1] > max) max = x[1];

        edges.push_back(Edge(x[0], x[1]));
        x.clear();
        arg1 >> x;
    }

    if (edges.size() == 0)
    {
        std::cerr << Isinf::argv0 << "no valid edges found" << std::endl;
        exit(4);
    }

    size_t i, fpat(0), npat(0);
    size_t nnod(0), fpos(0), fneg(0);
    size_t nedge(0), fagr(0), fdis(0);
    std::ifstream arg2(argv[2]);
    while (arg2.good())
    {
        // read target labels from argv[2]
        std::string t;
        getline(arg2, t);
        if (arg2.eof()) break;
        assert(!arg2.fail());
        assert(t.size() > max);
        assert(t[max] == '-');

        // read map state from standard input
        assert(std::cin.good());
        std::string s;
        getline(std::cin, s);
        assert(!std::cin.fail());
        assert(s.size() > max);
        assert(s[max] == '-'); 

        // count edge errors
        for (i = 0; i < edges.size(); ++i)
        {
            if ((s[edges[i].first] == s[edges[i].second]) &&
                (t[edges[i].first] != t[edges[i].second])) ++fagr;
            if ((s[edges[i].first] != s[edges[i].second]) &&
                (t[edges[i].first] == t[edges[i].second])) ++fdis;
            ++nedge;
        }

        // count node errors
        bool err = false;
        for (i = 0; i <= max; ++i)
        {
            assert(s[i] == '+' || s[i] == '-');
            assert(t[i] == '+' || t[i] == '-');
            if (s[i] == '+' && t[i] == '-')
            {
                ++fpos;
                err = true;
            }
            if (s[i] == '-' && t[i] == '+')
            {
                ++fneg;
                err = true;
            }
            ++nnod;
        }
        if (i < s.size())
            assert(s[i] != '+' && s[i] != '-');
        if (i < t.size())
            assert(t[i] != '+' && t[i] != '-');

        if (err) ++fpat;
        ++npat;
    }

    std::cout << std::fixed << std::setprecision(2);
    std::cout << "false edge agreement: " << std::setw(7) << fagr << '/'
        << nedge << " = " << std::setw(5) << (100.0*fagr)/nedge << "%\n";
    std::cout << "false edge disagrmnt: " << std::setw(7) << fdis << '/'
        << nedge << " = " << std::setw(5) << (100.0*fdis)/nedge << "%\n";
    std::cout << "false node positives: " << std::setw(7) << fpos << '/'
        << nnod << " = " << std::setw(5) << (100.0*fpos)/nnod << "%\n";
    std::cout << "false node negatives: " << std::setw(7) << fneg << '/'
        << nnod << " = " << std::setw(5) << (100.0*fneg)/nnod << "%\n";
    std::cout << "incorrect patterns:   " << std::setw(7) << fpat << '/'
        << npat << " = " << std::setw(5) << (100.0*fpat)/npat << "%\n";
}

