
// graph.hpp - embedded graph class for isinf package
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#ifndef GRAPH_HPP
#define GRAPH_HPP

#include <deque>
#include "utils.hpp"

namespace Isinf
{
    using namespace std;

    class Graph
    {
    public:
        class Edgelet
        {
        public:
            size_t node;
            size_t face;
            size_t index;
            bool agree, head, done;
            Edgelet* other;
            Edgelet* prev;
            Edgelet* next;

            Edgelet(void) : node(0), face(0), index(0),
                agree(false), head(false), done(false),
                other(0), prev(0), next(0) {};
        };

        class Vertex
        {
        public:
            bool state, done;
            Edgelet* edges;

            Vertex(void) : state(false), done(false), edges(0) {};
        };

        size_t nedges;
        vector<Vertex> nodes;
        vector<Edgelet*> edges;
    private:
        size_t nfaces;

        // generic breadth-first search:
        // invokes f on traversed edges
        // does not traverse edges marked as done
        // returns vector with node traversal order
        std::vector<size_t> bfs(size_t start, bool (*f)(Graph*, Edgelet*))
        {
            std::vector<size_t> q;
            q.reserve(nodes.size());
            q.push_back(start);
            nodes[start].done = true;
            size_t front = 0;

            while (front < q.size())
            {
                size_t i = q[front++];

                if (Edgelet* e = nodes[i].edges) do
                {
                    if (!(e->done || nodes[e->other->node].done))
                    {
                        if (f(this, e))
                        {
                            q.push_back(e->other->node);
                            nodes[e->other->node].done = true;
                        }
                    }
                    e = e->next;
                }
                while (e != nodes[i].edges);
            }

            return q;
        }

    public:
        ~Graph()
        {
            size_t i = edges.size();
            while (i)
            {
                if (edges[--i])
                {
                    delete edges[i]->other;
                    delete edges[i];
                }
            }
        }

        Graph(void) : nedges(0), nfaces(0) {};

        // build graph from embedding
        Graph(istream& is) : nedges(0), nfaces(0)
        {
            size_t i = 0, j, k;
            string s;
            Edgelet* e;

            do
            {
                getline(is, s);
                for (size_t n = 0; n < s.size(); n += k)
                {
                    if (sscanf(s.c_str() + n, " %zu %zn", &j, &k) < 1) continue;
                    e = new Edgelet;

                    e->index = j;
                    if (j >= edges.size()) edges.resize(j+1);
                    if (edges[j])
                    {
                        e->other = edges[j];
                        assert(!edges[j]->other);
                        edges[j]->other = e;
                        ++nedges;
                    }
                    else edges[j] = e;

                    e->node = i;
                    if (i >= nodes.size()) nodes.resize(i+1);
                    if (nodes[i].edges)
                    {
                        e->prev = nodes[i].edges->prev;
                        e->prev->next = e;
                        e->next = nodes[i].edges;
                        nodes[i].edges->prev = e;
                    }
                    else nodes[i].edges = e->prev = e->next = e;
                }
                ++i;
            }
            while (is.good());
            genus();
        }

        // compact the edge list, returns original indices
        vector<size_t>& compact(void)
        {
            vector<size_t>& idx = *new vector<size_t>;

            size_t j = 0;
            for (size_t i = 0; i < edges.size(); ++i)
            {
                if (edges[i])
                {
                    if (j < i)
                    {
                        edges[j] = edges[i];
                        edges[i]->index = edges[i]->other->index = j;
                    }
                    idx.push_back(i);
                    ++j;
                }
            }
            edges.resize(j);
            assert(edges.size() == nedges);
            return idx;
        }

        bool is_connected(void)
        {
            Disunion du(nodes.size());
            for (size_t i = 0; i < nedges; ++i)
                du.join(edges[i]->node, edges[i]->other->node);

            size_t root = du.find(0);
            bool c = true;
            for (size_t i = 0; c && i < nodes.size(); ++i)
                c &= (du.find(i) == root);

            return c;
        }

        // find nodes in same component as node i, store them in comp
        void component(size_t i, vector<size_t>& comp)
        {
            Disunion du(nodes.size());
            for (size_t j = 0; j < nedges; ++j)
                du.join(edges[j]->node, edges[j]->other->node);
            i = du.find(i);

            comp.clear();
            for (size_t j = 0; j < nodes.size(); j++)
                if (du.find(j) == i) comp.push_back(j);
        }

        bool add_edge(size_t i, size_t j)
        {
            // no self-links
            if (i == j) return false;

            // create nodes as necessary
            if (i >= nodes.size()) nodes.resize(i+1);
            if (j >= nodes.size()) nodes.resize(j+1);

            // check if edge exists already
            Edgelet* e1 = nodes[i].edges;
            Edgelet* e2;
            if (e1)
            {
                if (e1->other->node == j) return false;
                e2 = e1->next;
                while (e2 != e1)
                {
                    if (e2->other->node == j) return false;
                    e2 = e2->next;
                }
            }

            // create new edge
            e1 = new Edgelet;
            e2 = new Edgelet;

            e1->other = e2;
            e1->node = i;
            e1->index = nedges;
            if (nodes[i].edges)
            {
                e1->prev = nodes[i].edges->prev;
                e1->prev->next = e1;
                e1->next = nodes[i].edges;
                nodes[i].edges->prev = e1;
            }
            else nodes[i].edges = e1->prev = e1->next = e1;

            e2->other = e1;
            e2->node = j;
            e2->index = nedges;
            if (nodes[j].edges)
            {
                e2->prev = nodes[j].edges->prev;
                e2->prev->next = e2;
                e2->next = nodes[j].edges;
                nodes[j].edges->prev = e2;
            }
            else nodes[j].edges = e2->prev = e2->next = e2;

            edges.push_back(e1);
            ++nedges;
            nfaces = 0;
            return true;
        }

        // write out adjacency list, ignoring edgelets that are "done"
        void adj_list(ostream &o)
        {
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (!e->done) o << ' ' << e->index;
                    e = e->next;
                }
                while (e != nodes[n].edges);
                o << endl;
            }
        }

        // check if all edges are done
        bool all_done(void)
        {
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (!e->done) return false;
                    e = e->next;
                }
                while (e != nodes[n].edges);
            }
            return true;
        }

        // set done = false on nodes & edges
        void undone(void)
        {
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet* e = nodes[n].edges) do
                {
                    e->done = false;
                    e = e->next;
                }
                while (e != nodes[n].edges);
                nodes[n].done = false;
            }
        }

        // chords faces with large perimeter to prevent
        // quadratic growth of the expanded dual
        // returns the number of edges inserted
        // pentagons look to be the fastest
        size_t angulate(size_t gons = 5)
        {
            --gons;
            const size_t ne = nedges;

            undone();
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet *d = nodes[n].edges) do
                {
                    if (!d->done)
                    {
                        Edgelet *e = d->other->next;
                        size_t i = 2*gons;

                        while (i == 2*gons)
                        {
                            Edgelet *f = e, *g=0;

                            for (i = 1; i < 2*gons; ++i)
                            {
                                if (i <= gons) f->done = true;
                                f = f->other->next;
                                if (f == e) break;
                                if (i == gons) g = f;
                            }
                            if (i == 2*gons)   // insert chord (g,e)
                            {
                                Edgelet* b = new Edgelet;
                                b->index = nedges;
                                b->node = g->node;
                                b->prev = g->prev;
                                b->prev->next = b;
                                b->next = g;
                                g->prev = b;
                                b->done = true;

                                Edgelet* c = new Edgelet;
                                c->index = nedges;
                                c->node = e->node;
                                c->prev = e->prev;
                                c->prev->next = c;
                                c->next = e;
                                e->prev = c;

                                b->other = c;
                                c->other = b;
                                assert(edges.size() == nedges);
                                edges.push_back(b);
                                ++nedges;
                                ++nfaces;

                                e = g->other->next;
                            }
                            else if (i > gons)   // clean-up
                            {
                                for (f = g; f != e; --i, f = f->other->next)
                                    f->done = true;
                                assert(i == gons);
                            }
                        }
                    }
                    d = d->next;
                }
                while (d != nodes[n].edges);
            }
            assert(all_done());
            undone();
            return nedges - ne;
        }

        // count faces, return genus
        size_t genus(void)
        {
            if (nfaces == 0)
            {
                undone();
                for (size_t n = 0; n < nodes.size(); ++n)
                {
                    if (Edgelet* e = nodes[n].edges) do
                    {
                        if (!e->done)
                        {
                            Edgelet* f = e;
                            do
                            {
                                f->face = nfaces;
                                f->done = true;
                                f = f->other->next;
                            }
                            while (f != e);
                            ++nfaces;
                        }
                        e = e->next;
                    }
                    while (e != nodes[n].edges);
                }
                undone();
            }
            return ((nedges + 2) - (nodes.size() + nfaces))/2;
        }

        // try to improve state by flipping nodes
        // returns the achieved change in objective
        // NB: examines & changes state only on nodes, not edges
        double node_quench(const vector<double>& cost)
        {
            assert(std::numeric_limits<double>::has_infinity);

            // load priority queue
            PriorityVector<double> pv;
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                // compute value of flip
                double d = 0.0;
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (e->index < cost.size())
                        d += (nodes[e->node].state == nodes[e->other->node].state
                             ? cost[e->index] : -cost[e->index]);
                    e = e->next;
                }
                while (e != nodes[n].edges);
                pv[n] = d;
            }

            // and work it off
            double gain = 0.0;
            while (!pv.empty())
            {
                const size_t n = pv.top();
                if (isinf(double(pv[n]))) return gain;

                // flip best node
                gain += pv[n];
                pv[n] = -pv[n];
                nodes[n].state = !nodes[n].state;

                // re-cost neighbours
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (e->index < cost.size() &&
                        !isinf(double(pv[e->other->node])))
                        pv[e->other->node] +=
                            (nodes[e->node].state == nodes[e->other->node].state
                             ? 2*cost[e->index] : -2*cost[e->index]);
                    e = e->next;
                }
                while (e != nodes[n].edges);

                // disqualify losers
                if (pv.top() == n && pv[n] >= 0.0)
                    pv[n] = std::numeric_limits<double>::infinity();
            }

            assert(false);   // NOT REACHED
            return gain;
        }

        // implemented below
        bool make_odd(void);
        void mmc_nodestate(const vector<double>&);
        double edge_quench(const vector<double>&, bool);
    };

    // (used by make_odd, can't be member function)
    inline bool nop(Graph* g, Graph::Edgelet* e) { return true; };

    // create almost-odd edge orientation (non-recursive)
    bool Graph::make_odd(void)
    {
        bool odd = false;

        // find BFS traversal order
        undone();
        std::vector<size_t> o = bfs(0,  nop);
        undone();

        while (o.size())
        {
            // make odd in reverse order
            size_t n = o.back();
            o.pop_back();
            odd = false;
            Edgelet* last = 0;

            if (Edgelet* e = nodes[n].edges) do
            {
                if (nodes[e->other->node].done)
                {
                    assert(e->head ^ e->other->head);
                    if (e->head) odd = !odd;
                }
                else
                {
                    if (e->other->head)
                        assert(e->other->node == n);
                    else
                    {
                        e->head = true;
                        odd = !odd;
                        if (e->other->node != n) last = e;
                    }
                }
                e = e->next;
            }
            while (e != nodes[n].edges);

            nodes[n].done = true;

            if (!odd)  // reverse last edge
            {
                assert(last);
                assert(last->head);
                assert(!last->other->head);

                last->head = false;
                last->other->head = true;
                odd = !odd;
            }
        }

        undone();
        return odd;
    }

    // carry state from e->node to e->other->node
    // (used by mmc_nodestate & edge_quench, can't be member function)
    inline bool agree(Graph* g, Graph::Edgelet* e)
    {
        g->nodes[e->other->node].state = (g->nodes[e->node].state == e->agree);
        return true;
    }

    // find minimax-cost (MMC) node state given
    // edge costs & agreements, using spanning tree
    // NB: changes state only on nodes, not edges
    void Graph::mmc_nodestate(const vector<double>& cost)
    {
        assert(cost.size() <= nedges);
        undone();
        size_t i;

        // sort edges by cost
        vector<pair<double,size_t> > idx;
        for (i = 0; i < cost.size(); ++i)
            idx.push_back(pair<double,size_t>(-cost[i], i));
        for (; i < nedges; ++i)
            idx.push_back(pair<double,size_t>(0.0, i));
        sort(idx.begin(), idx.end());

        // build spanning tree (Kruskal's algorithm)
        Disunion du(nodes.size());
        for (i = 0; i < nedges; ++i)
        {
            Edgelet *e = edges[idx[i].second];
            if (!du.join(e->node, e->other->node))
            {
                e->done = true;
                e->other->done = true;
            }
            // else assert(e->index < cost.size());
        }

        // set node states along spanning tree
        bfs(nodes.size() - 1, agree);
    }

    // set back pointers to root of traversal
    // (used by edge_quench, can't be member function)
    inline bool orient(Graph* g, Graph::Edgelet* e)
    {
        g->nodes[e->other->node].edges = e->other;
        return true;
    }
    inline bool orient2(Graph* g, Graph::Edgelet* e)
    {
        if (g->nodes[e->other->node].edges == e->other)
            return false;
        else g->nodes[e->other->node].edges = e->other;
        return true;
    }

    // try to flip edges on the spanning tree
    // returns the achieved change in objective
    // NB: changes node states and edge agreements
    // does a very thorough job if slow = true
    double Graph::edge_quench(const vector<double>& cost, bool slow = false)
    {
        assert(std::numeric_limits<double>::has_infinity);

        size_t n;
        std::vector<double> delta(nedges, 0);
        vector<vector<size_t> > paths(cost.size());

        // need to use orient() once before orient2()
        for (n = 0; n < nodes.size(); ++n)
            nodes[n].done = false;
        for (n = 0; !edges[n]->done; ++n);
        bfs(edges[n]->node, orient);

        // compute cost of edge flips
        for (n = 0; n < cost.size(); ++n)
        {
            const double d = (nodes[edges[n]->node].state ==
                nodes[edges[n]->other->node].state ? cost[n] : -cost[n]);
            if (edges[n]->done)
            {
                for (size_t i = 0; i < nodes.size(); ++i)
                     nodes[i].done = false;
                bfs(edges[n]->node, orient2);

                for (size_t i = edges[n]->other->node;
                    i != edges[n]->node; i = nodes[i].edges->other->node)
                {
                    paths[n].push_back(nodes[i].edges->index);
                    delta[nodes[i].edges->index] += d;
                }
            }
            else delta[n] += d;
        }

        // load priority queue
        PriorityVector<double> pv;
        for (n = 0; n < nedges; ++n)
            if (!edges[n]->done) pv[n] = delta[n];
            else assert(delta[n] == 0);

        // and work it off
        double gain = 0;
        while (!pv.empty())
        {
            n = pv.top();
            if (!slow && pv[n] > 0) return gain;
            if (isinf(double(pv[n]))) return gain;

            // flip best edge
            gain += pv[n];
            edges[n]->agree = edges[n]->other->agree = !edges[n]->agree;
            edges[n]->done = true;

            // flip half the tree
            for (size_t i = 0; i < nodes.size(); ++i)
                 nodes[i].done = false;
            nodes[edges[n]->node].state = !nodes[edges[n]->node].state;
            if (bfs(edges[n]->node, agree).size() > nodes.size()/2)
                edges[n] = edges[n]->other;  // flip smaller half next time
            edges[n]->done = edges[n]->other->done = false;

            // re-cost everybody
            delta.clear();
            delta.resize(nedges, 0);
            delta[n] = (n < cost.size() ? 2*(nodes[edges[n]->node].state ==
                nodes[edges[n]->other->node].state ? cost[n] : -cost[n]) : 0);

            for (size_t i = 0; i < cost.size(); ++i)
            {
                if (edges[i]->done && (nodes[edges[i]->node].done
                    != nodes[edges[i]->other->node].done))
                {
                    const double d = 2*(nodes[edges[i]->node].state ==
                        nodes[edges[i]->other->node].state ? cost[i] : -cost[i]);

                    for (size_t k = 0; k < paths[i].size(); ++k)
                        delta[paths[i][k]] += d;
                }
            }

            // update priority queue
            for (size_t i = 0; i < nedges; ++i)
                if (!edges[i]->done && !isinf(double(pv[i])) && delta[i] != 0.0)
                    pv[i] += delta[i];

            // disqualify losers
            if (pv.top() == n && pv[n] >= 0.0)
                pv[n] = std::numeric_limits<double>::infinity();
        }

        assert(false);   // NOT REACHED
        return gain;
    };
}

#endif

