
// map.cpp - MAP state bia Blossom min-weight PM
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include <iomanip>
#include <fstream>
#include "blossom.hpp"

int main(int argc, char** argv)
{
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> "
            "<edgefeatures> < <params> > <nodestates>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];
    
    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read graph from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    Isinf::BlossomGraph g(arg1);
    assert(g.is_connected());
    if (g.genus() > 0)
        std::cerr << Isinf::argv0 << "warning: graph has genus "
            << g.genus() << std::endl;
    size_t n = g.angulate();

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(3);
    }

    while (arg2.good())
    {
        std::vector<double> e;

        while (e.size() < g.nedges - n)
        {
            std::vector<double> f;

            arg2 >> f;
            if (e.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << e.size() << std::endl;
                exit(4);
            }
            e.push_back(f*theta);
        }

        // find extended ground edge & MMC node state
        Isinf::CPUtimer *t = new Isinf::CPUtimer;
        g.call_blossom(e);

        /***
        // min-margin method:
        std::vector<double> ee(e);
        g.energy_margins(ee);
        g.mmc_nodestate(ee);

        // old method:
        for (size_t i = 0; i < e.size(); ++i)
            if (!g.edges[i]->agree) e[i] = -e[i];
        g.mmc_nodestate(e);
        for (size_t i = 0; i < e.size(); ++i)
            if (!g.edges[i]->agree) e[i] = -e[i];
        ***/

        // slack method:
        std::vector<double> slacks;
        for (size_t i = 0; i < g.nedges; ++i)
            slacks.push_back(g.pm->ExtendedSlack(i));
        // g.extended_slacks(slacks);  // DEPRECATED
        g.mmc_nodestate(slacks);

        delete t;

        // calculate & compare energies
        double lb = 0.0;
        double ub = 0.0;
        for (size_t i = 0; i < e.size(); ++i)
        {
            if (!g.edges[i]->agree) lb += e[i];
            if (g.nodes[g.edges[i]->node].state !=
                g.nodes[g.edges[i]->other->node].state) ub += e[i];
        }
        bool same = (std::abs(lb-ub) <= std::abs(lb)*1e-12);
        if (!g.genus()) assert(same);
        else assert(true || same || lb <= ub);

        // try quenching
        if (!same)
        {
std::cerr << std::setprecision(12) << ub << std::endl;
            t = new Isinf::CPUtimer("edge quenching");
            ub += g.edge_quench(e);
            delete t;
std::cerr << ub << std::endl;
            t = new Isinf::CPUtimer("node quenching");
            ub += g.node_quench(e);
            delete t;
std::cerr << ub << std::endl;
        }

        // report node state & energy
        for (size_t i = 0; i < g.nodes.size(); ++i)
            std::cout << (g.nodes[i].state != g.nodes.back().state ? '+' : '-');
        std::cout << ' ' << ub;
        if (!same) std::cout << ' ' << lb;
        std::cout << std::endl;
    }
}

