
// mme.cpp - perform max-margin CRF parameter estimation
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include <fstream>
#include <sstream>
#include "bt.hpp"
#include "blossom.hpp"

typedef struct
{
    std::vector<bool> labels;
    std::vector<std::vector<double> > features;
} Pattern;

std::vector<Pattern> train;

Isinf::BlossomGraph *graph;
double lambda = 1.0;

// given parameter vector x of length n,
// compute function value f and (sub)gradient vector g
void compfg(const int &n, const double x[], double &f, double g[])
{
    size_t i, j, k;
    std::vector<double> e;

    // regularization term
    f = 0.0;
    for (i = 0; i < (size_t) n; ++i)
    {
        f += x[i]*x[i];
        g[i] = train.size()*lambda*x[i];
    }
    f *= train.size()*lambda/2.0;

    for (i = 0; i < train.size(); ++i)
    {
        // compute edge energies
        e.clear();
        for (j = 0; j < train[i].features.size(); ++j)
        {
            double sum = 0.0;
            for (k = 0; k < (size_t) n; ++k)
                sum += x[k]*train[i].features[j][k];

            if (train[i].labels[graph->edges[j]->node] !=
                train[i].labels[graph->edges[j]->other->node])
            {
                for (k = 0; k < (size_t) n; ++k)
                    g[k] += train[i].features[j][k];
                sum += 1.0;
                f += sum;
            }
            else sum -= 1.0;

            e.push_back(sum);
        }

        // find worst margin violator
        graph->call_blossom(e);

        /*** PROBABLY NOT NEEDED
        // find corresponding node state
        // graph->mmc_nodestate(e);  // old method
        std::vector<double> slacks;
        for (j = 0; j < graph->nedges; ++j)
            slacks.push_back(graph->pm->ExtendedSlack(j));
        graph->mmc_nodestate(slacks);

        // set edge agreements accordingly
        for (j = 0; j < e.size(); ++j)
            graph->edges[j]->agree = (graph->nodes[graph->edges[j]->node].state
                == graph->nodes[graph->edges[j]->other->node].state);
        ***/

        // complete function/gradient calculation
        for (j = 0; j < train[i].features.size(); ++j)
        {
            if (!graph->edges[j]->agree)
            {
                f -= e[j];
                for (k = 0; k < (size_t) n; ++k)
                    g[k] -= train[i].features[j][k];
            }
        }
    }
}

int main(int argc, char** argv)
{
    if (argc < 3 || argc > 4)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> <edgefeatures>"
            " [<reg>] < <nodelabels>  > <params>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];

    // read graph from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    Isinf::BlossomGraph g(arg1);
    graph = &g;
    assert(graph->is_connected());
    assert(graph->genus() == 0);
    size_t n = graph->angulate();

    // read optional regularization parameter
    if (argc == 4)
    {
        std::istringstream arg3(argv[3]);
        arg3 >> lambda;
        if (arg3.fail() || lambda < 0.0)
        {
            std::cerr << "invalid regularization parameter"
                " (third argument)" << std::endl;
            exit(3);
        }
    }

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] <<
"'\n";
        exit(4);
    }

    std::vector<double> f;
    std::vector<std::vector<double> > e;
    Pattern p;
    size_t d = 0;

    while (arg2.good())
    {
        while (e.size() < graph->nedges - n)
        {
            arg2 >> f;
            if (e.empty() && f.empty()) break;

            if (d == 0) d = f.size();
            else if (f.size() != d)
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << d
                    << " features for edge " << e.size() << std::endl;
                exit(5);
            }
            e.push_back(f);
            f.clear();
        }
        if (e.empty()) break;
        p.features = e;
        e.clear();

        // read node labels from standard input
        size_t i;
        std::string s;
        getline(std::cin, s);
        assert(!std::cin.fail());
        assert(s.size() >= graph->nodes.size());

        for (i = 0; i < graph->nodes.size(); ++i)
        {
            assert(s[i] == '+' || s[i] == '-');
            p.labels.push_back(s[i] == '+');
        }
        if (i < s.size())
            assert(s[i] != '+' && s[i] != '-');

        train.push_back(p);
        p.labels.clear();
    }

    // perform the optimisation
    std::vector<double> x(d, 0.0);
    Isinf::CPUtimer *t = new Isinf::CPUtimer;
    call_bt(x);
    delete t;
    std::cout << x << std::endl;
}

