
// rpm.cpp - MAP state for non-planar graphs via reweighting
// (c) 2009 Nic Schraudolph - see COPYRIGHT.txt for details
//
// Very early & hacky implementation - expect numerical problems!
// As described in:
// N. N. Schraudolph.  Polynomial-Time Exact Inference in NP-Hard
// Binary MRFs via Reweighted Perfect Matching. In Proc. 13th Intl.
// Conf. Artificial Intelligence and Statistics (AIstats),
// pp. 717–724, Journal of Machine Learning Research, 2010.

#include <iomanip>
#include <fstream>
#include "bt.hpp"
#include "blossom.hpp"

std::vector<Isinf::BlossomGraph*> graph;
std::vector<double> xbar;
std::vector<size_t> idx;
std::vector<size_t> ne;

// given parameter vector x of length n,
// compute function value f and (sub)gradient vector g
void compfg(const int &n, const double x[], double &f, double g[])
{
    f = 0;

    // evaluate graph[0]
    std::vector<double> xx(xbar);
    for (size_t k = 0; int(k) < n; ++k)
        xx[idx[k]] -= x[k];
    graph[0]->call_blossom(xx);
    for (size_t j = 0; j < ne[0]; ++j)
    {
        if (!graph[0]->edges[j]->agree) f -= xx[j];
        // if (graph[0]->pm->SureEdge(j))
            // xx[j] += (graph[0]->edges[j]->agree ? 1e3 : -1e3);
    }

    // evaluate other graphs
    size_t k = 0;
    size_t dis = 0;
    for (size_t i = 1; i < graph.size(); ++i)
    {
        graph[i]->call_blossom(&x[k], ne[i]);
        for (size_t j = 0; j < ne[i]; ++j, ++k)
        {
            if (!graph[i]->edges[j]->agree) f -= x[k];
            if ((g[k] = graph[i]->edges[j]->agree -
                graph[0]->edges[idx[k]]->agree)) ++dis;
        }
    }
    assert(int(k) == n);

std::cerr << dis << '/' << n << ' '
  << std::setprecision(12) << -f << std::endl;
}

int main(int argc, char** argv)
{
    if (argc < 3)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> [<embedding> [...]]"
            " <edgefeatures> < <params> > <nodestates>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];
    
    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read the embeddings
    int narg;
    for (narg = 1; narg < argc - 1; ++narg)
    {
        std::ifstream arg(argv[narg]);
        if (arg.fail())
        {
            std::cerr << Isinf::argv0 << "cannot open file '" << argv[narg] << "'\n";
            exit(2);
        }

        graph.push_back(new Isinf::BlossomGraph(arg));
        std::vector<size_t>& tmp = graph.back()->compact();
        if (narg > 1)
        {
            idx.insert(idx.end(), tmp.begin(), tmp.end());

            // assert identical node numbers
            for (size_t i = 0; i < graph.back()->nedges; ++i)
            {
                size_t a = graph.back()->edges[i]->node;
                size_t b = graph.back()->edges[i]->other->node;
                if (a > b) std::swap(a, b);
                size_t j = graph.front()->edges[tmp[i]]->node;
                size_t k = graph.front()->edges[tmp[i]]->other->node;
                if (j > k) std::swap(j, k);
                assert(a == j);
                assert(b == k);
            }
        }
        delete &tmp;
        assert(graph.back()->is_connected());
        ne.push_back(graph.back()->nedges);
        graph.back()->angulate();
    }

    // count edge occurrences
    std::vector<size_t> nn;
    nn.resize(ne[0], 1);
    for (size_t k = 0; k < idx.size(); ++k) ++nn[idx[k]];

    // features are last argument
    std::ifstream arg2(argv[narg]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0
            << "cannot open file '" << argv[narg] << "'\n";
        exit(3);
    }

    while (arg2.good())
    {
        // read features
        while (xbar.size() < ne[0])
        {
            std::vector<double> feat;

            arg2 >> feat;
            if (xbar.empty() && feat.empty()) exit(0);

            if (feat.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << feat.size()
                    << " instead of " << theta.size()
                    << " features for edge " << xbar.size() << std::endl;
                exit(4);
            }
            xbar.push_back(feat*theta);
            feat.clear();
        }

        // upper bound MAP energy
        double ub = 0.0;
        graph[0]->call_blossom(xbar);
        std::vector<double> slacks;
        for (size_t i = 0; i < xbar.size(); ++i)
            slacks.push_back(-graph[0]->pm->ExtendedSlack(i));
        graph[0]->mmc_nodestate(slacks);

        for (size_t j = 0; j < xbar.size(); ++j)
        {
            if (graph[0]->nodes[graph[0]->edges[j]->node].state !=
                graph[0]->nodes[graph[0]->edges[j]->other->node].state)
                ub += xbar[j];
        }
        ub += graph[0]->edge_quench(xbar);
        ub += graph[0]->node_quench(xbar);

        // initial configuration
        std::vector<double> x;
        for (size_t k = 0; k < idx.size(); ++k)
            x.push_back(xbar[idx[k]]/double(nn[idx[k]]));

        // optimisation
        Isinf::CPUtimer *t = new Isinf::CPUtimer;
        double nrg = ub;
        double lb = nrg - 2.0;
        while (nrg > lb + 0.999999)
        {
            lb = -call_bt(x, -nrg);
            std::vector<double> xx(xbar);
            for (size_t k = 0; k < x.size(); ++k)
                xx[idx[k]] -= x[k];
            graph[0]->call_blossom(xx);

            // slacks.clear();
            // for (size_t i = 0; i < xbar.size(); ++i)
                // slacks.push_back(-graph[0]->pm->ExtendedSlack(i));
            graph[0]->mmc_nodestate(xbar);

            nrg = 0.0;
            for (size_t j = 0; j < xbar.size(); ++j)
            {
                if (graph[0]->nodes[graph[0]->edges[j]->node].state !=
                    graph[0]->nodes[graph[0]->edges[j]->other->node].state)
                    nrg += xbar[j];
            }
            if (nrg > lb + 0.999999)
                nrg += graph[0]->edge_quench(xbar);
            if (nrg > lb + 0.999999)
                nrg += graph[0]->node_quench(xbar);
        }
        delete t;

        // and report it
        for (size_t j = 0; j < graph[0]->nodes.size(); ++j)
            std::cout << (graph[0]->nodes[j].state
                != graph[0]->nodes.back().state ? '+' : '-');
        std::cout << std::setprecision(12) << ' '
           //  << ub << ' '
            << nrg
            // << ' ' << lb
            << std::endl;

        xbar.clear();
        x.clear();
    }

    // clean-up
    for (size_t i = 0; i < graph.size(); ++i) delete graph[i];
}

