
// utils.hpp - miscellaneous utilities for isinf package
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#ifndef UTILS_HPP
#define UTILS_HPP

#include <cmath>
#include <assert.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <set>

namespace Isinf
{
    // disjoint union forest data structure
    // with union by size & path compression

    class Disunion
    {
    private:
        size_t components;
        std::vector<bool> noroot;
        std::vector<size_t> thesize;
        std::vector<size_t> parent;

    public:
        Disunion(size_t n) : components(n), noroot(n, false), thesize(n, 1)
        {
            for (size_t i = 0; i < n; ++i)
                parent.push_back(i);
        }

        inline size_t comps(void) const { return components; }
        inline size_t size(void)  const { return thesize.size(); }
        inline size_t size(size_t i )   { return thesize[find(i)]; }

        inline size_t find(size_t i)
        {
            size_t j, root = i;

            while (noroot[root]) root = parent[root];

            // path compression
            if (noroot[i])
                while (noroot[parent[i]])
                {
                    j = parent[i];
                    parent[i] = root;
                    i = j;
                }

            return parent[root];
        }

        inline bool join(size_t i, size_t j)
        {
            i = find(i);
            j = find(j);
            if (i == j) return false;

            // union by size
            if (thesize[i] < thesize[j])
            {
                parent[i] = j;
                noroot[i] = true;
                thesize[j] += thesize[i];
            }
            else
            {
                parent[j] = i;
                noroot[j] = true;
                thesize[i] += thesize[j];
            }

            --components;
            return true;
        }

        // relabels roots to consecutive integers
        inline void compact(void)
        {
            size_t k = 0;
            for (size_t i = 0; i < size(); ++i)
                if (!noroot[i]) parent[i] = k++;
            assert(k == components);
        }
    };


    // an object that stores argv[0] for error messages

    class Argv0
    {
    private:
        std::string argv0;

    public:
        inline Argv0() : argv0("") {};

        inline Argv0& operator=(char *a0)
        {
#ifdef ARGV0
            std::string s(a0);
            size_t pos(s.rfind('/'));
            argv0 = (pos == std::string::npos ? s : s.substr(++pos));
            argv0 += ": ";
#endif
            return *this;
        }

        inline friend std::ostream& operator<<(std::ostream& os, Argv0& me)
        {
            os << me.argv0;
            return os;
        }
    };
    static Argv0 argv0;  // should be set to argv[0] by each program


#ifdef CPUTIME
#include <sys/resource.h>

    // an object that reports its own lifetime in CPU seconds

    class CPUtimer
    {
    private:
        struct rusage start;
        std::string str;

    public:
        inline CPUtimer(const std::string &s = "") : str(s)
        {
            getrusage(RUSAGE_SELF, &start);
        }

        inline ~CPUtimer(void)
        {
            struct rusage r;
            getrusage(RUSAGE_SELF, &r);

            double user = r.ru_utime.tv_usec - start.ru_utime.tv_usec;
            user /= 1e6;
            user += r.ru_utime.tv_sec - start.ru_utime.tv_sec;

            double sys = r.ru_stime.tv_usec - start.ru_stime.tv_usec;
            sys /= 1e6;
            sys += r.ru_stime.tv_sec - start.ru_stime.tv_sec;

            std::cerr << argv0 << str << ' ' << user << " + " << sys << " s\n";
        }
    };
#else
    class CPUtimer {};
#endif


    // priority queue with indexed elements, implemented via std::set
    // resolution of ties: among all elements with lowest value, top()
    // returns the one with lowest index, rtop() a pseudo-random one.

    template<class T = double> class PriorityVector
    {
    private:
        std::vector<T> priority;
        typedef std::pair<T, size_t> payload;
        std::set<payload> queue;
        size_t index;
        bool indexed;

    public:
        const T end(void) const { return std::numeric_limits<T>::max(); }
        bool empty(void) const { assert(!indexed); return queue.empty(); }
        size_t size(void) const { assert(!indexed); return queue.size(); }
        size_t top(void) const { assert(!indexed); return queue.begin()->second; }

        PriorityVector(void) : indexed(false) {};

        void pop(void)  // remove top element
        {
            assert(!indexed);
            if (empty()) return;
            queue.erase(queue.begin());
        }

        size_t rtop(void) const
        {
            assert(!indexed);

            typename std::set<payload>::iterator p = queue.begin();
            const T best(p->first);
            std::vector<size_t> cands;
            cands.push_back(p->second);

            while (++p != queue.end())
            {
                if (p->first != best) break;
                cands.push_back(p->second);
            }

            return cands[rand()%cands.size()];
        }

        PriorityVector& operator[](const size_t i)
        {
            index = i;
            indexed = true;
            return *this;
        }

        void erase(void)  // remove element with given index
        {
            assert(indexed);
            indexed = false;
            assert(index < priority.size());
            assert(priority[index] < end());
            queue.erase(payload(priority[index], index));
        }

        operator const T(void) // deliver rvalue
        {
            assert(indexed);
            indexed = false;
            return index < priority.size() ? priority[index] : end();
        }

        const T& operator=(const T& value)
        {
            assert(indexed);
            indexed = false;
            if (index >= priority.size())
                priority.resize(index + 1, end());
            else queue.erase(payload(priority[index], index));

            priority[index] = value;
            queue.insert(payload(value, index));
            return value;
        }

        const T& operator+=(const T& value)
        {
            erase();
            priority[index] += value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator-=(const T& value)
        {
            erase();
            priority[index] -= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator*=(const T& value)
        {
            erase();
            priority[index] *= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator/=(const T& value)
        {
            erase();
            priority[index] /= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator%=(const T& value)
        {
            erase();
            priority[index] %= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator<<=(const T& value)
        {
            erase();
            priority[index] <<= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator>>=(const T& value)
        {
            erase();
            priority[index] >>= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator&=(const T& value)
        {
            erase();
            priority[index] &= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator|=(const T& value)
        {
            erase();
            priority[index] |= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator^=(const T& value)
        {
            erase();
            priority[index] ^= value;
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator++()
        {
            erase();
            ++priority[index];
            queue.insert(payload(priority[index], index));
            return priority[index];
        }

        const T& operator--()
        {
            erase();
            --priority[index];
            queue.insert(payload(priority[index], index));
            return priority[index];
        }
    };
}


// operators & formatted I/O for vectors

// inner product
template <class T>
inline T operator*(const std::vector<T>& a, const std::vector<T>& b)
{
    assert(a.size() == b.size());
    T sum(0);
    for (size_t i = 0; i < a.size(); ++i)
        sum += a[i]*b[i];
    return sum;
}

template <class T>
inline std::vector<T>& operator+=(std::vector<T>& a, const T& b)
{
    typename std::vector<T>::iterator i(a.begin());
    while (i != a.end()) *i++ += b;
    return a;
}

template <class T>
inline std::vector<T>& operator-=(std::vector<T>& a, const T& b)
{
    typename std::vector<T>::iterator i(a.begin());
    while (i != a.end()) *i++ -= b;
    return a;
}

template <class T>
inline std::vector<T>& operator*=(std::vector<T>& a, const T& b)
{
    typename std::vector<T>::iterator i(a.begin());
    while (i != a.end()) *i++ *= b;
    return a;
}

template <class T>
inline std::vector<T>& operator/=(std::vector<T>& a, const T& b)
{
    typename std::vector<T>::iterator i(a.begin());
    while (i != a.end()) *i++ /= b;
    return a;
}

template <class T>
inline std::vector<T>& operator+=(std::vector<T>& a, const std::vector<T>& b)
{
    assert(a.size() == b.size());
    typename std::vector<T>::iterator i(a.begin());
    typename std::vector<T>::const_iterator j(b.begin());
    while (i != a.end()) *i++ += *j++;
    return a;
}

template <class T>
inline std::vector<T>& operator-=(std::vector<T>& a, const std::vector<T>& b)
{
    assert(a.size() == b.size());
    typename std::vector<T>::iterator i(a.begin());
    typename std::vector<T>::const_iterator j(b.begin());
    while (i != a.end()) *i++ -= *j++;
    return a;
}

template <class T>
inline std::vector<T>& operator*=(std::vector<T>& a, const std::vector<T>& b)
{
    assert(a.size() == b.size());
    typename std::vector<T>::iterator i(a.begin());
    typename std::vector<T>::const_iterator j(b.begin());
    while (i != a.end()) *i++ *= *j++;
    return a;
}

template <class T>
inline std::vector<T>& operator/=(std::vector<T>& a, const std::vector<T>& b)
{
    assert(a.size() == b.size());
    typename std::vector<T>::iterator i(a.begin());
    typename std::vector<T>::const_iterator j(b.begin());
    while (i != a.end()) *i++ /= *j++;
    return a;
}

template <class T>
inline std::ostream& operator<<(std::ostream& os, const std::vector<T>& x)
{
    typename std::vector<T>::const_iterator i(x.begin());
    while(i != x.end()) os << *i++ << ' ';
    return os;
}

template <class T>
inline std::istream& operator>>(std::istream& is, std::vector<T>& x)
{
    std::string s;
    const size_t n = x.size();

    while (x.size() == n)
    {
        getline(is, s);
        if (is.fail()) break;

        std::istringstream iss(s);
        T item;

        iss >> item;
        while (iss.good())
        {
            x.push_back(item);
            iss >> item;
        }
        if (!iss.fail())
            x.push_back(item);
    }

    return is;
}

#endif

