
// velim.cpp - brute-force calculation of MAP states,
// log-partition function, and edge marginal probabilities
// (uses variable elimination in order of node indices)
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include <string>
#include <fstream>
#include <iomanip>
#include "utils.hpp"

inline double logsumexp(double a, double b)
{
    if (b > a) std::swap(a, b);
    return a + log1p(exp(b-a));
}

typedef std::pair<std::pair<size_t, size_t>, size_t> Edge;

std::vector<Edge> edges;
std::vector<bool> state;
std::vector<double> energy;
std::vector<double> margs;
std::vector<std::string> mapstates;
// std::vector<double> probs;

double lnZ = log(0.0);
double min = 1e9999999;

void z(size_t n, size_t i, double e)
{
    size_t j;
    double tmp;

    if (n == state.size())
    {
        // report progress?
        bool progress = true;
        for (j = state.size()/2; j < state.size(); ++j)
            if (state[j])
            {
                progress = false;
                break;
            }

        if (progress)
        {
            size_t fact = 1;
            size_t d = 0;
            for (j = 0; j < state.size()/2; ++j)
            {
                if (state[state.size()/2 - j - 1]) d += fact;
                fact <<= 1;
            }
            j = state.size()/6;
            if (j > 2)
            std::cerr << std::setw(j) << std::setprecision(j-3) << std::fixed
                << (200*d)/double(fact) << "% done               \r";
        }

        // partition function
        lnZ = logsumexp(lnZ, -e);
        // probs.push_back(-e);

        // edge marginals
        for (j = 0; j < edges.size(); ++j)
            if (state[edges[j].first.first] != state[edges[j].first.second])
                margs[edges[j].second] = logsumexp(margs[edges[j].second], -e);

        // MAP states
        if (e < min)
        {
            min = e;
            mapstates.resize(0);
        }
        if (e == min)
        {
            std::string s = *new std::string;
            for (size_t k = 0; k < state.size(); ++k)
                s += (state[k] == state.back() ? '-' : '+');
            mapstates.push_back(s);
        }
        return;
    }

    tmp = e;
    state[n] = false;
    for (j = i; j < edges.size(); ++j)
    {
        if (edges[j].first.first != n) break;
        if (state[edges[j].first.first] != state[edges[j].first.second])
            tmp += energy[edges[j].second];
    }
    z(n + 1, j, tmp);

    tmp = e;
    state[n] = true;
    for (j = i; j < edges.size(); ++j)
    {
        if (edges[j].first.first != n) break;
        if (state[edges[j].first.first] != state[edges[j].first.second])
            tmp += energy[edges[j].second];
    }
    z(n + 1, j, tmp);

    return;
}


int main(int argc, char** argv)
{
    Isinf::argv0 = argv[0];
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0]
            << " <edgelist> <edgefeatures> < <params>" << std::endl;
        exit(1);
    }

    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read edge list from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    std::vector<size_t> x;
    size_t nodes = 0;
    size_t i = 0;

    arg1 >> x;
    while (x.size() >= 2)
    {
        if (x[0] == x[1])  // self-loop!
        {
            std::cerr << Isinf::argv0 << "error at node " << x[0]
                << ": self-loops not allowed" << std::endl;
            exit(3);
        }

        if (x[0] < x[1]) std::swap(x[0], x[1]);
        if (x[0] > nodes) nodes = x[0];

        edges.push_back(Edge(std::pair<size_t,size_t>(x[0], x[1]), i++));
        x.clear();
        arg1 >> x;
    }
    ++nodes;

    if (edges.size() == 0)
    {
        std::cerr << Isinf::argv0 << "no valid edges found" << std::endl;
        exit(3);
    }

    // elimination order given by node index
    std::sort(edges.begin(), edges.end());
    state.resize(nodes);
    state[0] = false;

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(4);
    }

    while (arg2.good())
    {
        while (energy.size() < edges.size())
        {
            std::vector<double> f;

            arg2 >> f;
            if (energy.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << energy.size() << std::endl;
                exit(5);
            }
            energy.push_back(f*theta);
        }

        margs.resize(edges.size(), log(0.0));

        // use variable elimination
        Isinf::CPUtimer *t = new Isinf::CPUtimer;
        z(1, 0, 0.0);
        std::cerr.copyfmt(std::cout);
        std::cerr << "                                 \r";
        delete t;
        energy.clear();

        // report MAP states
        for (size_t k = 0; k < mapstates.size(); ++k)
            std::cout << mapstates[k] << ' ';
        std::cout << min << std::endl;
        mapstates.clear();

// DEBUG: get edge state for 1st MAP state
/***
std::vector<bool> foo(edges.size(), false);
for (size_t k = 0; k < edges.size(); ++k)
     foo[edges[k].second] = (mapstates[0][edges[k].first.first] ==
         mapstates[0][edges[k].first.second]);
for (size_t k = 0; k < edges.size(); ++k)
    std::cout << (foo[k] ? '+' : '-');
std::cout << std::endl;
***/

        // report edge marginals
        for (size_t k = 0; k < margs.size(); ++k)
            std::cout << exp(margs[k] - lnZ) << ' ';
        std::cout << std::endl;
        margs.clear();

        // report log partition function
        std::cerr << lnZ << std::endl;
        // for (int k = 0; k < probs.size(); ++k)
            // std::cout << exp(probs[k] - lnZ) << std::endl;
        lnZ = log(0.0);
        min = 1e9999999;
    }
}

